@@ -1280,3 +1280,66 @@ fn s2n_strerror_name () {
     assert!(!ptr.is_null());
 }
 
+#[test]
+#[cfg(feature = "internal")]
+fn s2n_config_client_hello_cb_enable_poll () {
+    let ptr = crate::s2n_config_client_hello_cb_enable_poll as *const ();
+    assert!(!ptr.is_null());
+}
+
+#[test]
+#[cfg(feature = "internal")]
+fn s2n_connection_get_config () {
+    let ptr = crate::s2n_connection_get_config as *const ();
+    assert!(!ptr.is_null());
+}
+
+#[test]
+#[cfg(feature = "quic")]
+fn s2n_config_enable_quic () {
+    let ptr = crate::s2n_config_enable_quic as *const ();
+    assert!(!ptr.is_null());
+}
+
+#[test]
+#[cfg(feature = "quic")]
+fn s2n_connection_enable_quic () {
+    let ptr = crate::s2n_connection_enable_quic as *const ();
+    assert!(!ptr.is_null());
+}
+
+#[test]
+#[cfg(feature = "quic")]
+fn s2n_connection_get_quic_transport_parameters () {
+    let ptr = crate::s2n_connection_get_quic_transport_parameters as *const ();
+    assert!(!ptr.is_null());
+}
+
+#[test]
+#[cfg(feature = "quic")]
+fn s2n_connection_is_quic_enabled () {
+    let ptr = crate::s2n_connection_is_quic_enabled as *const ();
+    assert!(!ptr.is_null());
+}
+
+#[test]
+#[cfg(feature = "quic")]
+fn s2n_connection_set_quic_transport_parameters () {
+    let ptr = crate::s2n_connection_set_quic_transport_parameters as *const ();
+    assert!(!ptr.is_null());
+}
+
+#[test]
+#[cfg(feature = "quic")]
+fn s2n_connection_set_secret_callback () {
+    let ptr = crate::s2n_connection_set_secret_callback as *const ();
+    assert!(!ptr.is_null());
+}
+
+#[test]
+#[cfg(feature = "quic")]
+fn s2n_error_get_alert () {
+    let ptr = crate::s2n_error_get_alert as *const ();
+    assert!(!ptr.is_null());
+}
+
